/* simple NFA representation
 * pesco 2021-2022, isc license
 */

#include <limits.h>	/* UCHAR_MAX, INT_MAX, INT_MIN */
#include <stdlib.h>	/* size_t */

#if UCHAR_MAX >= INT_MAX	/* it could happen */
#error need INT_MAX > UCHAR_MAX
#endif

/*
 * Note: INT_MAX > UCHAR_MAX implies sizeof(int) > sizeof(char), and since
 * sizeof(char) = 1, the number of bits in an int is n * CHAR_BIT with n > 1.
 * In other words, int has at least CHAR_BIT bits more than char. CHAR_BIT is
 * greater than 1 (actually, at least 8), so it follows that the number
 * INT_MAX / 2 is still greater than UCHAR_MAX (by at least 7 bits).
 */
#define END	(INT_MAX / 2)	/* special input symbol meaning end of input */
#define OUTPUT	(END + 1)	/* start of output range */
#define INNER	INT_MIN		/* start of "inner input" range */

struct range {
	int base;
	int n;			/* interval length; negative = inverted */
};
struct state {
	struct range chr;
	unsigned int edge_chr;	/* advance (positive offset) only */
	int edge_eps;		/* can point back (negative offset) */
};
struct nfa {
	struct state *state;
	int size;		/* allocated size, i.e. num. of states - 1 */
};
typedef struct nfa NFA;


/* elementary NFAs */
NFA fail(void);
NFA range(unsigned char, unsigned char);
NFA nrange(unsigned char, unsigned char);
NFA chr(unsigned char);
NFA any(void);
NFA symbol(int);		/* input only */
NFA output(int);
NFA outchr(unsigned char);
NFA end(void);
NFA nop(void);
NFA think(int);
NFA epsilon(void);

/* combinators */
NFA str(const char *);
NFA seq_(NFA, ...);		/* -> macro seq() */
NFA choice_(NFA, ...);		/* -> macro choice() */
NFA opt(NFA);
NFA rep(int, NFA);
NFA plus(NFA);
NFA star(NFA);
#define seq(...)	seq_(__VA_ARGS__, (NFA){NULL, -1})
#define choice(...)	choice_(__VA_ARGS__, (NFA){NULL, -1})


/* running NFAs */

typedef int sub_t(int symbol, size_t state, int pos, void *env);
typedef int out_t(struct range r, size_t state, int pos, void *env);

struct prep;
struct prep *nfaprep(NFA, sub_t *, out_t *, void *);
void nfareset(struct prep *);
int nfarun(struct prep *, const char *, size_t);
int nfacont(struct prep *, const char *, size_t);
int nfaend(struct prep *);

/* convenience wrapper combining nfaprep, nfarun, and nfaend */
int nfamatch(NFA, sub_t *, out_t *, void *, const char *, size_t);


/* diagnostics */
void nfaprint(NFA);
